@@ -1,5 +1,4 @@
 #define COPYRIGHT_NOTICE "Copyright (C) 2003 Free Software Foundation, Inc."
-#define BUG_ADDRESS "bonzini@gnu.org"
 
 /*  GNU SED, a batch stream editor.
     Copyright (C) 1989,90,91,92,93,94,95,98,99,2002,2003
@@ -7,7 +6,7 @@
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2, or (at your option)
+    the Free Software Foundation; either version 3, or (at your option)
     any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -59,6 +58,8 @@
 # define ATOI(x)	strtoul(x, NULL, 0)
 #endif
 
+char *program_name;
+
 int extended_regexp_flags = 0;
 
 /* If set, fflush(stdout) on every line output. */
@@ -70,9 +71,15 @@ bool no_default_output = false;
 /* If set, reset line counts on every new file. */
 bool separate_files = false;
 
+/* If set, follow symlinks when processing in place */
+bool follow_symlinks = false;
+
 /* How do we edit files in-place? (we don't if NULL) */
 char *in_place_extension = NULL;
 
+/* The mode to use to read files, either "rt" or "rb".  */
+char *read_mode = "rt";
+
 /* Do we need to be pedantically POSIX compliant? */
 enum posixicity_types posixicity;
 
@@ -82,6 +89,25 @@ countT lcmd_out_line_len = 70;
 /* The complete compiled SED program that we are going to run: */
 static struct vector *the_program = NULL;
 
+static void usage P_((int));
+static void
+contact(errmsg)
+  int errmsg;
+{
+  FILE *out = errmsg ? stderr : stdout;
+#ifndef REG_PERL
+  fprintf(out, _("GNU sed home page: <http://www.gnu.org/software/sed/>.\n\
+General help using GNU software: <http://www.gnu.org/gethelp/>.\n"));
+#endif
+
+  /* Only print the bug report address for `sed --help', otherwise we'll
+     get reports for other people's bugs.  */
+  if (!errmsg)
+    fprintf(out, _("E-mail bug reports to: <%s>.\n\
+Be sure to include the word ``%s'' somewhere in the ``Subject:'' field.\n"),
+	  PACKAGE_BUGREPORT, PACKAGE);
+}
+
 static void usage P_((int));
 static void
 usage(status)
@@ -105,15 +131,25 @@ Usage: %s [OPTION]... {script-only-if-no-other-script} [input-file]...\n\
                  add the script to the commands to be executed\n"));
   fprintf(out, _("  -f script-file, --file=script-file\n\
                  add the contents of script-file to the commands to be executed\n"));
+#ifdef ENABLE_FOLLOW_SYMLINKS
+  fprintf(out, _("  --follow-symlinks\n\
+                 follow symlinks when processing in place\n"));
+#endif
   fprintf(out, _("  -i[SUFFIX], --in-place[=SUFFIX]\n\
                  edit files in place (makes backup if extension supplied)\n"));
+#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__) || defined(MSDOS) || defined(__EMX__)
+  fprintf(out, _("  -b, --binary\n\
+                 open files in binary mode (CR+LFs are not processed specially)\n"));
+#endif
   fprintf(out, _("  -l N, --line-length=N\n\
                  specify the desired line-wrap length for the `l' command\n"));
   fprintf(out, _("  --posix\n\
                  disable all GNU extensions.\n"));
   fprintf(out, _("  -r, --regexp-extended\n\
                  use extended regular expressions in the script.\n"));
+#ifdef REG_PERL
   fprintf(out, PERL_HELP);
+#endif
   fprintf(out, _("  -s, --separate\n\
                  consider files as separate rather than as a single continuous\n\
                  long stream.\n"));
@@ -128,9 +164,7 @@ non-option argument is taken as the sed script to interpret.  All\n\
 remaining arguments are names of input files; if no input files are\n\
 specified, then the standard input is read.\n\
 \n"));
-  fprintf(out, _("E-mail bug reports to: %s .\n\
-Be sure to include the word ``%s'' somewhere in the ``Subject:'' field.\n"),
-	  BUG_ADDRESS, PACKAGE);
+  contact (status);
 
   ck_fclose (NULL);
   exit (status);
@@ -142,12 +176,13 @@ main(argc, argv)
   char **argv;
 {
 #ifdef REG_PERL
-#define SHORTOPTS "snrRue:f:l:i::V:"
+#define SHORTOPTS "bsnrRuEe:f:l:i::V:"
 #else
-#define SHORTOPTS "snrue:f:l:i::V:"
+#define SHORTOPTS "bsnruEe:f:l:i::V:"
 #endif
 
   static struct option longopts[] = {
+    {"binary", 0, NULL, 'b'},
     {"regexp-extended", 0, NULL, 'r'},
 #ifdef REG_PERL
     {"regexp-perl", 0, NULL, 'R'},
@@ -163,6 +198,9 @@ main(argc, argv)
     {"unbuffered", 0, NULL, 'u'},
     {"version", 0, NULL, 'v'},
     {"help", 0, NULL, 'h'},
+#ifdef ENABLE_FOLLOW_SYMLINKS
+    {"follow-symlinks", 0, NULL, 'F'},
+#endif
     {NULL, 0, NULL, 0}
   };
 
@@ -170,6 +208,7 @@ main(argc, argv)
   int return_code;
   const char *cols = getenv("COLS");
 
+  program_name = argv[0];
   initialize_main (&argc, &argv);
 #if HAVE_SETLOCALE
   /* Set locale according to user's wishes.  */
@@ -215,6 +254,10 @@ main(argc, argv)
 	  the_program = compile_file(the_program, optarg);
 	  break;
 
+	case 'F':
+	  follow_symlinks = true;
+	  break;
+
 	case 'i':
 	  separate_files = true;
 	  if (optarg == NULL)
@@ -241,6 +284,12 @@ main(argc, argv)
 	  posixicity = POSIXLY_BASIC;
 	  break;
 
+        case 'b':
+	  read_mode = "rb";
+	  break;
+
+	/* Undocumented, for compatibility with BSD sed.  */
+	case 'E':
 	case 'r':
 	  if (extended_regexp_flags)
 	    usage(4);
@@ -275,6 +324,9 @@ This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE,\n\
 to the extent permitted by law.\n\
 "), COPYRIGHT_NOTICE);
+	  fputc('\n', stdout);
+	  contact(false);
+
 	  ck_fclose (NULL);
 	  exit (0);
 	case 'h':
